/*
 * vic web/gif interface
 *
 * Robert Olson
 * Argonne National Laboratory
 * olson@mcs.anl.gov
 * (c) 1996
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <errno.h>

extern "C" char *sys_errlist[];

#include "decoder.h"
#include "renderer.h"
#include "device-output.h"
#include "vw.h"
#include "color-pseudo.h"

extern "C" {
#define GD_EXTERNAL
#include "gd.h"
#include "faketcl.h"
void gdImageGifChannel(gdImagePtr im, membuf);
};

#define DEBUG

extern "C" u_char quant_cmap[];

#define UCLIMIT(x) ((t = (x)), (t &= ~(t>>31)), (t | ~((t-256) >> 31)))

class DitherColorModel : public PseudoColorModel {
public:
	virtual int alloc_colors();
	virtual int command(int argc, const char*const* argv);
};

class DitherWindowRenderer;

typedef void (DitherWindowRenderer::*DitherMethod)(const u_char*,
						   u_int, u_int,
						   u_int, u_int) const;

class DitherWindowRenderer : public PseudoWindowRenderer {
public:
	DitherWindowRenderer(VideoWindow* vw, int decimation,
			     DitherColorModel& cm);
	void render(const u_char* frm, int off, int x, int w, int h);

protected:
	DitherColorModel& cm_;
	virtual void update();
	virtual void disable();
	DitherMethod method_;
	void dither_422(const u_char* frm, u_int off, u_int x,
			u_int width, u_int height) const;
	void dither_down2_422(const u_char* frm, u_int off, u_int x,
			      u_int width, u_int height) const;
	void dither_down4_422(const u_char* frm, u_int off, u_int x,
			      u_int width, u_int height) const;
	void dither_down_422(const u_char* frm, u_int off, u_int x,
			     u_int width, u_int height) const;
	void dither_up2_422(const u_char* frm, u_int off, u_int x,
			    u_int width, u_int height) const;
	void dither_411(const u_char* frm, u_int off, u_int x,
			u_int width, u_int height) const;
	void dither_down2_411(const u_char* frm, u_int off, u_int x,
			      u_int width, u_int height) const;
	void dither_down4_411(const u_char* frm, u_int off, u_int x,
			      u_int width, u_int height) const;
	void dither_down_411(const u_char* frm, u_int off, u_int x,
			     u_int width, u_int height) const;
	void dither_up2_411(const u_char* frm, u_int off, u_int x,
			    u_int width, u_int height) const;
};


class ClientRequest
{
public:
    ClientRequest();
    ClientRequest(const char *chan_name, const char *img_type, int one_shot,
		  const char *separator = 0);

    void send_image(membuf);
    void send_oneshot(membuf);
    void send_stream(membuf);

    inline int one_shot() { return one_shot_; };
    inline int writable() { return writable_; };
    inline void set_writable(int v) { writable_ = v; };

    ClientRequest *next;
    ClientRequest *prev;

    int writable_;
    
    char *channel_name_;
    char *img_type_;
    char *separator_;
    int one_shot_;
    Tcl_Channel tcl_channel_;
    Tcl_File tcl_file_;
    int fd_;
    FILE *fp_;
};

class ClientRequestPool
{
public:
    void enqueue(ClientRequest *);
    void dequeue(ClientRequest *);

    void print_queue(const char *msg);

    void send_image(membuf);

    void writable(const char *conn);

protected:
    void dequeue_oneshots();

    ClientRequest head;
};

class GifOutputDevice : public OutputDevice
{
public:
    GifOutputDevice(const char *name, const char *dir);

    virtual int command(int argc, const char*const* argv);

    void initialize();

    void dump(u_char *buf, int w, int h);
    void resize(int w, int h);

protected:
    ClientRequestPool requests;
    int initialized_;
    const char *output_directory_;
    VideoImage *image_buffer_;
    VideoWindow *video_window_;
    DitherColorModel *color_model_;
    int dither_to_gif_map_[256];
    gdImagePtr  gd_image_;
    membuf membuf_;
};

class GifWindowRenderer : public DitherWindowRenderer
{
public:
    GifWindowRenderer(VideoWindow* vw, int decimation,
		      DitherColorModel& cm, GifOutputDevice *dev) :
	DitherWindowRenderer(vw, decimation, cm),
	device(dev)
    {
	setcolor(1);
    }

    virtual void resize(int w, int h);
    void dump();
protected:
    GifOutputDevice *device;
};

void GifWindowRenderer::resize(int w, int h)
{
	int outw = outw_;
	int outh = outh_;
	compute_scale(w, h);
	/*
	 * Doing a resize can change the output window size
	 * (but it's not likely because this size is mainly
	 * dependent on the size of the window).
	 */
	if (outw != outw_ || outh != outh_) {
	}
	doupdate();
	pixbuf_ = new u_char[w * h];
	device->resize(outw_, outh_);
}

void GifWindowRenderer::dump()
{
//    printf("renderer dump...\n");
    device->dump(pixbuf_, outw_, outh_);
}

#if 0
void GifWindowRenderer::dump()
{
    static int done = 0;
    static int map[256];
    
    if (done)
	return;

    done = 1;

    FILE *fp;
    const ColorModel::color *c;

//    printf("dumping frame %dx%d\n", outw_, outh_);
    fp = fopen("/tmp/out.ppm", "w");
    fprintf(fp, "P6\n%d %d\n255\n", outw_, outh_);
    for (int i = 0; i < outw_ * outh_; i++)
    {
	c = cm_.lookup(pixbuf_[i]);
	u_char pix[3];
	pix[0] = c->r;
	pix[1] = c->g;
	pix[2] = c->b;
	    
	fwrite(pix, 3, 1, fp);
    }
    fclose(fp);

    fp = fopen("/tmp/out.gif", "w");

    gdImagePtr img;

    img = gdImageCreate(outw_, outh_);

    for (int i = 0; i < 256; i++)
    {
	c = cm_.lookup(i);
	map[i] = gdImageColorAllocate(img, c->r, c->g, c->b);
/*
	img->open[i] = 0;
	img->red[i] = c->r;
	img->green[i] = c->g;
	img->blue[i] = c->b;
*/
    }
//    img->colorsTotal = 160;
    for (int i = 0; i < outw_ * outh_; i++)
    {
	gdImageSetPixel(img, i % outw_, i / outw_, map[pixbuf_[i]]);
    }
    gdImageGif(img, fp);
    fclose(fp);
    gdImageDestroy(img);
}
#endif

class GifOutboard : public Renderer
{
public:
    GifOutboard(GifOutputDevice &device, int format);
    ~GifOutboard();
    
    virtual int command(int argc, const char*const* argv);

    virtual void setcolor(int c) {}

 protected:
    GifOutputDevice &device_;
    int scale_;

    void resize(int width, int height);
};

class GifOutboard_yuv : public GifOutboard
{
public:
    GifOutboard_yuv(GifOutputDevice &device, int format)
	: GifOutboard(device, format) {};

protected:
    virtual int consume(const VideoFrame*);
    virtual void yuv_to_rgb(unsigned char *frame,
			    unsigned long ts) = 0;
};

class GifOutboard_422 : public GifOutboard_yuv
{
public:
    GifOutboard_422(GifOutputDevice &device);

protected:
    virtual void yuv_to_rgb(unsigned char *frame,
			    unsigned long ts);
};

class GifOutboard_411 : public GifOutboard_yuv
{
public:
    GifOutboard_411(GifOutputDevice &device);

protected:
    virtual void yuv_to_rgb(unsigned char *frame, unsigned long ts);
};

class GifOutboard_ren : public GifOutboard
{
public:
    GifOutboard_ren(GifOutputDevice &device, GifWindowRenderer *r, int format);

protected:
    virtual int consume(const VideoFrame*);
    virtual void resize(int w, int h);
    GifWindowRenderer *renderer_;
};

static GifOutputDevice shmem_device("gif", "/tmp/gifs");

class GifVideoImage : public VideoImage {
public:
    GifVideoImage(Tk_Window tk, int width, int height)
	: VideoImage(tk, width, height)
    {
	pixbuf_  = new u_char[width * height];
    };

    virtual ~GifVideoImage()
    {
	delete pixbuf_;
    };

    void putimage(Display *, Window, GC, int sx, int sy, int x, int y,
		  int w, int h) const;
    
    inline u_char* pixbuf() { return pixbuf_; }
protected:
    u_char *pixbuf_;
};

void GifVideoImage::putimage(Display *, Window, GC,
			     int sx, int sy, int x, int y, int w, int h) const
{
    printf("gif putimage\n");
}

ClientRequest::ClientRequest(const char *chan_name, const char *img_type,
			     int one_shot, const char *separator)
{
    int mode, type;
    Tcl& tcl = Tcl::instance();

    channel_name_ = new char[strlen(chan_name) + 1];
    strcpy(channel_name_, chan_name);

    img_type_ = new char[strlen(img_type) + 1];
    strcpy(img_type_, img_type);

    if (separator)
    {
	separator_ = new char[strlen(separator) + 5];
	sprintf(separator_, "\n--%s\n", separator);
    }
    else
	separator_ = NULL;

    one_shot_ = one_shot;

    tcl_channel_ = Tcl_GetChannel(tcl.interp(), channel_name_, &mode);
	    
//    printf("got tclchan=%x mode=%d\n", tcl_channel_, mode);

    /*

    tcl_file_ = Tcl_GetChannelFile(tcl_channel_, TCL_WRITABLE);
	    
    fd_ = (int) Tcl_GetFileInfo(tcl_file_, &type);

//    printf("fd=%d type=%d\n", fd_, type);

    fp_ = fdopen(fd_, "w");
    */

    next = prev = NULL;
}

ClientRequest::ClientRequest() :
    channel_name_(NULL),
    img_type_(NULL),
    separator_(NULL),
    tcl_channel_(NULL),
    tcl_file_(NULL),
    one_shot_(-1),
    fd_(-1),
    fp_(NULL),
    next(NULL),
    prev(NULL)
{

}

void ClientRequest::send_image(membuf mb)
{
    if (one_shot_)
    {
	send_oneshot(mb);
    }
    else
    {
	send_stream(mb);
    }
}

void ClientRequest::send_oneshot(membuf mb)
{
    Tcl& tcl = Tcl::instance();
    char buf[200];
    sprintf(buf, "Content-length: %d\n\n", membuf_len(mb));
/*
    gdImageGif(img, fp_);
    fclose(fp_);
*/
    if (Tcl_Write(tcl_channel_, buf, strlen(buf)) < 0)
	printf("Tcl write error: %d %s\n", Tcl_GetErrno(), sys_errlist[Tcl_GetErrno()]);
    int n = Tcl_Write(tcl_channel_, membuf_buf(mb), membuf_len(mb));
//    printf("tcl_write %d returns %d of %d\n", getpid(), n, membuf_len(mb));
    Tcl_Flush(tcl_channel_);
    Tcl_UnregisterChannel(tcl.interp(), tcl_channel_);

/*
    char fbuf[1000];
    FILE *fp;
    sprintf(fbuf, "/tmp/out.%d", getpid());
    fp = fopen(fbuf, "w");
    fwrite(membuf_buf(mb), membuf_len(mb), 1, fp);
    fclose(fp);
    */
}

/*
 * Send a server push stream:
 *  Content-type: image/gif
 *  <blank>
 *  <image>
 *  --separator
 */
#define BAIL do { printf("Stream closing\n"); one_shot_ = 1; return; } while (0);
void ClientRequest::send_stream(membuf mb)
{
    if (!writable())
    {
	printf("stream %s not writable\n", channel_name_);
	return;
    }
    
    char buf[400];
    sprintf(buf, "Content-type: image/gif\n\n");
    sprintf(buf, "Content-type: image/gif\nContent-length: %d\n\n", membuf_len(mb));
    if (Tcl_Write(tcl_channel_, buf, strlen(buf)) < 0)
	BAIL;

    if (Tcl_Write(tcl_channel_, membuf_buf(mb), membuf_len(mb)) < 0)
	BAIL;

    if (Tcl_Write(tcl_channel_, separator_, strlen(separator_)) < 0)
	BAIL;
	
//    if (Tcl_Flush(tcl_channel_) != TCL_OK)
//	BAIL;
    set_writable(0);
}

void ClientRequestPool::print_queue(const char *msg)
{
    ClientRequest *c;
    
    printf("Request pool: %s\n", msg);

    printf("Head: %x next=%x prev=%x\n", &head, head.next, head.prev);
    for (c = head.next; c != NULL; c = c->next)
    {
	printf("   %x: name=%s next=%x prev=%x one_shot_=%d\n",
	       c, c->channel_name_, c->next, c->prev, c->one_shot_);
    }
}


void ClientRequestPool::enqueue(ClientRequest *req)
{
//    print_queue("before enqueue");
    req->next = head.next;
    req->prev = &head;
    if (head.next)
	head.next->prev = req;
    head.next = req;
//    print_queue("after enqueue");
}

void ClientRequestPool::dequeue(ClientRequest *req)
{
//    print_queue("before dequeue");
    req->prev->next = req->next;
    if (req->next)
	req->next->prev = req->prev;
//    print_queue("after dequeue");
}

void ClientRequestPool::dequeue_oneshots()
{
    ClientRequest *c, *next_c;

    for (c = head.next; c != NULL; c = next_c)
    {
	next_c = c->next;
	if (c->one_shot())
	{
	    c->prev->next = c->next;
	    if (c->next)
		c->next->prev = c->prev;
	    delete c;
	}
    }
}

void ClientRequestPool::send_image(membuf mb)
{
    ClientRequest *c;

//    print_queue("before send_image");
    for (c = head.next; c != NULL; c = c->next)
    {
	c->send_image(mb);
    }
    dequeue_oneshots();
}

void ClientRequestPool::writable(const char *conn)
{
    ClientRequest *c;

    for (c = head.next; c != NULL; c = c->next)
    {
	if (strcmp(c->channel_name_, conn) == 0)
	{
	    c->set_writable(1);
	    return;
	}
    }
}

GifOutputDevice::GifOutputDevice(const char *name, const char *dir) :
			OutputDevice(name), output_directory_(dir),
			initialized_(0), gd_image_(NULL)
{
    printf("create gifoutput for %s\n", output_directory_);
    membuf_ = membuf_new(500000);
}

void GifOutputDevice::initialize()
{
    if (initialized_)
	return;
    initialized_ = 1;
    /*
     * Brace yourself: create a video window and ditherer...
     */

    signal(SIGPIPE, SIG_IGN);
    
    char *cm;
    Tcl& tcl = Tcl::instance();
    Tk_Window mainWindow = Tk_MainWindow(tcl.interp());

    printf("have main = %x\n", mainWindow);

    video_window_ = new VideoWindow(".gif");

    printf("created window %x\n", video_window_);

    image_buffer_ = new GifVideoImage(mainWindow, 352, 288);

    video_window_->setsize(352, 288);

    video_window_->setimage(image_buffer_);

    /*
    color_model_ = new DitherColorModel();

    tcl.evalf("setvis %s", color_model_->name());
    
    color_model_->alloc_colors();
    */
/*    cm = tcl.var("V(colorModel)");*/
    cm = Tcl_GetVar2(tcl.interp(), "V", "colorModel", TCL_GLOBAL_ONLY);
    printf("got cm=%s\n", cm);
    color_model_ = (DitherColorModel *) TclObject::lookup(cm);
    printf("got color_model=%x %s\n", color_model_, color_model_->name());

}

void GifOutputDevice::resize(int w, int h)
{
    if (gd_image_)
    {
	gdImageDestroy(gd_image_);
	gd_image_ = NULL;
    }

    printf("create image %dx%d\n", w, h);
    gd_image_ = gdImageCreate(w, h);

    for (int i = 0; i < 256; i++)
    {
	const ColorModel::color *c;
	
	c = color_model_->lookup(i);
	dither_to_gif_map_[i] = gdImageColorAllocate(gd_image_,
						     c->r, c->g, c->b);
    }
}

void GifOutputDevice::dump(u_char *buf, int w, int h)
{
    FILE *fp;
    static done = 0;
    ClientRequest *c;
    static int frames = 0;
    static double next_send;
    static double start_time;
    double now;

    /*
     * If it's not first time thru and it's not time to send...
     */
    now = Decoder::gettimeofday();
    if (frames > 0 && now  < next_send)
    {
	return;
    }

    if (frames == 0)
	start_time = now;
    frames++;
/*
    printf("frames=%d time=%f fps: %5.2f\n",
	   frames, now - start_time,
	   (double) frames * 1e6 / (now - start_time));
*/
    
//    printf("device dump %dx%d\n", w, h);

    for (int i = 0; i < w * h; i++)
    {
//	gdImageSetPixel(gd_image_, i % w, i / w,
//			dither_to_gif_map_[buf[i]]);
	gd_image_->pixels[i%w][i/w] = dither_to_gif_map_[buf[i]];
    }

    int white = gdImageColorAllocate(gd_image_, 255, 255, 255);
    gdImageLine(gd_image_, w / 2, 0, w/2, h, white);
    gdImageLine(gd_image_, 0, h/2, w, h / 2, white);

    membuf_reset(membuf_);
    gdImageGifChannel(gd_image_, membuf_);

    requests.send_image(membuf_);
/*
    requests.send_image(gd_image_);
*/

    next_send = now + (1e6 / 5.0);

    if (done)
	return;
    done = 1;
    fp = fopen("/tmp/out.gif", "w");
    fwrite(membuf_buf(membuf_), membuf_len(membuf_), 1, fp);
    fclose(fp);
/*
    gdImageGif(gd_image_, fp);
    fclose(fp);
*/    
}

int GifOutputDevice::command(int argc, const char*const* argv)
{
    Tcl& tcl = Tcl::instance();
#if defined(DEBUG) && 0
    printf("GifOutputDevice::command argc=%d ", argc);
    for (int i = 0; i < argc; i++)
	printf("\"%s\" ", argv[i]);
    printf("\n");
#endif /* DEBUG */

    if (argc == 3) {
	if (strcmp(argv[1], "assistor") == 0)
	{
#ifdef FT_RGB
	    if (strcmp(argv[2], "rgb") == 0)
	    {
		printf("Assist with rgb\n");
		GifOutboard *p = new GifOutboard_rgb(*this);
		tcl.result(p->name());
		return (TCL_OK);
	    }
#endif
	}
	else if (strcmp(argv[1], "renderer") == 0) {
	    if (strcmp(argv[2], "422") == 0) {
		initialize();

		GifWindowRenderer *r = new GifWindowRenderer(video_window_,
							     422,
							     *color_model_,
							     this);
		printf("created renderer %x %s\n", r, r->name());
		GifOutboard* p = new GifOutboard_ren(*this, r, FT_YUV_422);
		tcl.result(p->name());
		return (TCL_OK);
	    }

	    if (strcmp(argv[2], "411") == 0) {
		initialize();
		GifWindowRenderer *r = new GifWindowRenderer(video_window_,
							     411,
							     *color_model_,
							     this);
		printf("created renderer %x %s\n", r, r->name());
		
		GifOutboard* p = new GifOutboard_ren(*this, r, FT_YUV_411);
		tcl.result(p->name());
		return (TCL_OK);
	    }
	}
	else if (strcmp(argv[1], "writable") == 0)
	{
	    requests.writable(argv[2]);
	    return (TCL_OK);
	}
    }
    else if (argc == 4)
    {
	if (strcmp(argv[1], "request_oneshot") == 0)
	{
	    const char *img_type;
	    const char *conn;

	    img_type = argv[2];
	    conn = argv[3];

//	    printf("Request_oneshot: img_type='%s' conn='%s'\n",
//		   img_type, conn);

	    ClientRequest *req;

	    req = new ClientRequest(conn, img_type, 1);
	    requests.enqueue(req);

	    return(TCL_OK);
	}
    }
    else if (argc == 5)
    {
	if (strcmp(argv[1], "request_stream") == 0)
	{
	    const char *img_type;
	    const char *conn;
	    const char *separator;

	    img_type = argv[2];
	    conn = argv[3];
	    separator = argv[4];

//	    printf("Request_stream: img_type='%s' conn='%s' sep='%s'\n",
//		   img_type, conn, separator);

	    ClientRequest *req;

	    req = new ClientRequest(conn, img_type, 0, separator);
	    requests.enqueue(req);

	    return(TCL_OK);
	}
    }

    return (OutputDevice::command(argc, argv));
}

GifOutboard::GifOutboard(GifOutputDevice &device, int format) :
    				Renderer(format), device_(device)
{
    printf("Construct GifOutboard for format %d\n", format);
}

GifOutboard::~GifOutboard()
{
    printf("Destruct GifOutboard()\n");
}

int GifOutboard::command(int argc, const char*const* argv)
{
    Tcl& tcl = Tcl::instance();
#ifdef DEBUG
    printf("GifOutboard::command argc=%d ", argc);
    for (int i = 0; i < argc; i++)
	printf("\"%s\" ", argv[i]);
    printf("\n");
#endif /* DEBUG */

    if (argc == 3) {
	if (strcmp(argv[1], "scale") == 0) {
	    int s = atoi(argv[2]);
	    if (scale_ != s) {
		scale_ = s;
		resize(width_, height_);
	    }
	    return (TCL_OK);
	}
/*
	if (strcmp(argv[1], "port") == 0) {
	    port_ = device_.lookup_port(argv[2]);
	    startup_vl();
	    return (TCL_OK);
	}
*/
    }
    return (Renderer::command(argc, argv));
}

int GifOutboard_yuv::consume(const VideoFrame* frame)
{
    if (!samesize(frame))
	resize(frame->width_, frame->height_);
    
    yuv_to_rgb(frame->bp_, frame->ts_);
	       
    return 0;
}

void GifOutboard::resize(int w, int h)
{
/*
    image_->width = TEX_WIDTH;
    image_->height = TEX_HEIGHT;
*/
    
    width_ = w;
    height_ = h;
    framesize_ = w * h;
}

#ifdef FT_RGB
GifOutboard_rgb::GifOutboard_rgb(GifOutputDevice &device) :
    				GifOutboard(device, FT_RGB)
{
    printf("Construct GifOutboard_rgb\n");
}

int GifOutboard_rgb::consume(const VideoFrame* frame)
{
    static int framect = 0;
    RGBFrame *rgb = (RGBFrame *) frame;
/*
    printf("Consume rgb %dx%d ts=%d len=%d\n", frame->width_, frame->height_,
	frame->ts_, rgb->len_);
*/
    
    memcpy(image_->data, frame->bp_, rgb->len_);
    image_->timestamp = framect++;
	  
    return 1;
}
#endif

GifOutboard_422::GifOutboard_422(GifOutputDevice &device) :
    				GifOutboard_yuv(device, FT_YUV_422)
{
    printf("Construct GifOutboard_422\n");
}

void GifOutboard_422::yuv_to_rgb(unsigned char *frame, unsigned long ts)
{
    unsigned char *yp, *up, *vp;
    unsigned char *yl, *ul, *vl;
    int w, h, ui, vi, yi, t;
    int framesize;
    int y, u, v;
    unsigned char r, g, b;
    static int once = 0;
    
    yp = frame;
    up = frame + framesize_;
    vp = up + framesize_ / 2;

    yi = ui = vi = 0;

    yl = yp;
    ul = up;
    vl = vp;

    for (h = 0; h < height_; h++)
    {
	yl = yp + h * width_;
	ul = up + h * width_ / 2;
	vl = vp + h * width_ / 2;
	for (w = 0; w < width_; w += 2)
	{
	    u = (int) (*ul++ - 128);
	    v = (int) (*vl++ - 128);
	  
	    y = (int) (*yl++);

	    b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	    g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	    r = (unsigned char) UCLIMIT(y + 385 * v / 256);

/*
	    image_->data[TEX_OFFSET_R(w, h)] = r;
	    image_->data[TEX_OFFSET_G(w, h)] = g;
	    image_->data[TEX_OFFSET_B(w, h)] = b;
 */

	    y = (int) (*yl++);

	    b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	    g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	    r = (unsigned char) UCLIMIT(y + 385 * v / 256);

/*
	    image_->data[TEX_OFFSET_R(w + 1, h)] = r;
	    image_->data[TEX_OFFSET_G(w + 1, h)] = g;
	    image_->data[TEX_OFFSET_B(w + 1, h)] = b;
	    */
	}
    }
/*
    printf("New frame ts=%d %dx%d\n", image_->timestamp,
	   image_->width, image_->height);
*/

    printf("New frame ts=%d %dx%d\n", ts, width_, height_);

#if 0
    if (once == 0)
    {
	once = 1;
	FILE *fp = fopen("img.ppm", "w");
	if (fp == NULL)
	{
	    printf("can't open img.ppm\n");
	}
	else
	{
	    fprintf(fp, "P6\n");
	    fprintf(fp, "%d %d\n", TEX_WIDTH, TEX_HEIGHT);
	    fprintf(fp, "255\n");
	    fwrite(image_->data, 3, TEX_WIDTH * TEX_HEIGHT, fp);
	    fclose(fp);
	}
    }
#endif
}

GifOutboard_411::GifOutboard_411(GifOutputDevice &device) :
    				GifOutboard_yuv(device, FT_YUV_411)
{
    printf("Construct GifOutboard_411\n");
}

void GifOutboard_411::yuv_to_rgb(unsigned char *frame, unsigned long ts)
{
    unsigned char *yp, *up, *vp;
    int x, w, h, ui, vi, yi, k, t;
    int framesize;
    double y, u, v;
    unsigned char r, g, b;
    static int once = 0;
    
    yp = frame;
    up = frame + framesize_;
    vp = up + framesize_ / 2;

    yi = ui = vi = 0;

/*
    image_->width = TEX_WIDTH;
    image_->height = TEX_HEIGHT;
    */
    
    for (h = 0; h < height_; h++)
    {
	x = 0;
	for (w = 0; w < width_; w += 2)
	{
/*
	    u = (double) (up[ui++] - 128.0);
	    v = (double) (vp[vi++] - 128.0);
	    */

	    for (k = 0; k < 2; k++)
	    {
		r = g = b = (yp[yi++]);

		if (w % 5 != 0)
		{
/*
		    if (x < TEX_WIDTH)
		    {
			image_->data[TEX_OFFSET_R(x, h)] = r;
			image_->data[TEX_OFFSET_G(x, h)] = g;
			image_->data[TEX_OFFSET_B(x, h)] = b;
		    }
*/
		    x++;
		}
	    }
	}
    }
/*    image_->timestamp = ts;*/
/*    printf("New frame ts=%d %dx%d\n", image_->timestamp,
	   image_->width, image_->height);*/
    printf("New frame ts=%d %dx%d\n", ts, width_, height_);
#if 0
    if (once == 0)
    {
	once = 1;
	FILE *fp = fopen("img.ppm", "w");
	if (fp == NULL)
	{
	    printf("can't open img.ppm\n");
	}
	else
	{
	    fprintf(fp, "P6\n");
	    fprintf(fp, "%d %d\n", TEX_WIDTH, TEX_HEIGHT);
	    fprintf(fp, "255\n");
	    fwrite(image_->data, 3, TEX_WIDTH * TEX_HEIGHT, fp);
	    fclose(fp);
	}
    }
#endif
}

GifOutboard_ren::GifOutboard_ren(GifOutputDevice &device,
				 GifWindowRenderer *r, int format) :
    				GifOutboard(device, format),
				renderer_(r)
{
    printf("Construct GifOutboard_ren renderer=%x\n");
}

void GifOutboard_ren::resize(int w, int h)
{
    printf("Resize outboard to %dx%d\n", w, h);
    width_ = w;
    height_ = h;
    framesize_ = w * h;

    renderer_->resize(w, h);
}

int GifOutboard_ren::consume(const VideoFrame *frame)
{
/*
    printf("Consume ren %dx%d ts=%d len=%d\n", frame->width_, frame->height_,
	frame->ts_);
*/
    
    if (!samesize(frame))
	resize(frame->width_, frame->height_);

    renderer_->render(frame->bp_, 0, 0, frame->width_, frame->height_);
    renderer_->dump();
    return 1;
}
